@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	28 October 1990
- * Last Edited:	1 April 1997
+ * Last Edited:	9 October 1999
  *
- * Copyright 1997 by the University of Washington
+ * Copyright 1999 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -42,10 +42,13 @@
 #include <ctype.h>
 #include <errno.h>
 extern int errno;		/* just in case */
+#include <signal.h>
 #include "misc.h"
 
 
 /* Autologout timer */
+#define KODTIMEOUT 60*5
+#define LOGINTIMEOUT 60*3
 #define TIMEOUT 60*30
 
 
@@ -64,10 +67,11 @@ extern int errno;		/* just in case */
 
 /* Global storage */
 
-char *version = "3.34";		/* server version */
+char *version = "4.54";		/* server version */
 short state = LISN;		/* server state */
 short critical = NIL;		/* non-zero if in critical code */
 MAILSTREAM *stream = NIL;	/* mailbox stream */
+long idletime = 0;		/* time we went idle */
 unsigned long nmsgs = 0;	/* number of messages */
 unsigned long current = 1;	/* current message number */
 unsigned long size = 0;		/* size of current message */
@@ -79,7 +83,7 @@ unsigned long *msg = NIL;	/* message translation vector */
 
 /* Function prototypes */
 
-void main (int argc,char *argv[]);
+int main (int argc,char *argv[]);
 void clkint ();
 void kodint ();
 void hupint ();
@@ -94,41 +98,49 @@ short c_nack (char *t);
 
 /* Main program */
 
-void main (int argc,char *argv[])
+int main (int argc,char *argv[])
 {
   char *s,*t;
   char cmdbuf[TMPLEN];
 #include "linkage.c"
-  openlog ("ipop2d",LOG_PID,LOG_MAIL);
+				/* initialize server */
+  server_init (argv[0],"pop",NIL,"pop",clkint,kodint,hupint,trmint);
   /* There are reports of POP2 clients which get upset if anything appears
    * between the "+" and the "POP2" in the greeting.
    */
-  printf ("+ POP2 %s v%s server ready\015\012",mylocalhost (),version);
+  printf ("+ POP2 %s v%s server ready\015\012",tcp_serverhost (),version);
   fflush (stdout);		/* dump output buffer */
-				/* set up traps */
-  server_traps (clkint,kodint,hupint,trmint);
   state = AUTH;			/* initial server state */
   while (state != DONE) {	/* command processing loop */
-    alarm (TIMEOUT);		/* get a command under timeout */
+    idletime = time (0);	/* get a command under timeout */
+    alarm ((state != AUTH) ? TIMEOUT : LOGINTIMEOUT);
+    clearerr (stdin);		/* clear stdin errors */
     while (!fgets (cmdbuf,TMPLEN-1,stdin)) {
-      if (errno==EINTR) errno=0;/* ignore if some interrupt */
+      if (ferror (stdin) && (errno == EINTR)) clearerr (stdin);
       else {
-	syslog (LOG_INFO,
-		"Connection broken while reading line user=%.80s host=%.80s",
-		user ? user : "???",tcp_clienthost (cmdbuf));
+	char *e = ferror (stdin) ?
+	  strerror (errno) : "Command stream end of file";
+	alarm (0);		/* disable all interrupts */
+	server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
+	syslog (LOG_INFO,"%s while reading line user=%.80s host=%.80s",
+		e,user ? user : "???",tcp_clienthost ());
+	state = DONE;
 	mail_close (stream);	/* try to close the stream gracefully */
 	stream = NIL;
 	_exit (1);
       }
     }
     alarm (0);			/* make sure timeout disabled */
+    idletime = 0;		/* no longer idle */
 				/* find end of line */
     if (!strchr (cmdbuf,'\012')) {
-      puts ("- Command line too long\015");
+      server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
+      fputs ("- Command line too long\015\012",stdout);
       state = DONE;
     }
     else if (!(s = strtok (cmdbuf," \015\012"))) {
-      puts ("- Missing or null command\015");
+      server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
+      fputs ("- Missing or null command\015\012",stdout);
       state = DONE;
     }
     else {			/* dispatch based on command */
@@ -137,7 +149,7 @@ void main (int argc,char *argv[])
       t = strtok (NIL,"\015\012");
       if ((state == AUTH) && !strcmp (s,"HELO")) state = c_helo (t,argc,argv);
       else if ((state == MBOX || state == ITEM) && !strcmp (s,"FOLD"))
-	state = c_fold (t);
+ 	state = c_fold (t);
       else if ((state == MBOX || state == ITEM) && !strcmp (s,"READ"))
 	state = c_read (t);
       else if ((state == ITEM) && !strcmp (s,"RETR")) state = c_retr (t);
@@ -146,25 +158,29 @@ void main (int argc,char *argv[])
       else if ((state == NEXT) && !strcmp (s,"NACK")) state = c_nack (t);
       else if ((state == AUTH || state == MBOX || state == ITEM) &&
 	       !strcmp (s,"QUIT")) {
-	if (t) puts ("- Bogus argument given to QUIT\015");
-	else {			/* valid sayonara */
-				/* expunge the stream */
-	  if (stream && nmsgs) mail_expunge (stream);
-	  puts ("+ Sayonara\015");/* acknowledge the command */
-	}
+	server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
 	state = DONE;		/* done in either case */
+	if (t) fputs ("- Bogus argument given to QUIT\015\012",stdout);
+	else {			/* expunge the stream */
+	  if (stream && nmsgs) stream = mail_close_full (stream,CL_EXPUNGE);
+	  stream = NIL;		/* don't repeat it */
+				/* acknowledge the command */
+	  fputs ("+ Sayonara\015\012",stdout);
+	}
       }
       else {			/* some other or inappropriate command */
+	server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
 	printf ("- Bogus or out of sequence command - %s\015\012",s);
 	state = DONE;
       }
     }
     fflush (stdout);		/* make sure output blatted */
   }
-  mail_close (stream);		/* clean up the stream */
-  syslog (LOG_INFO,"Logout user %.80s host %.80s",user ? user : "???",
-	  tcp_clienthost (cmdbuf));
-  exit (0);			/* all done */
+				/* clean up the stream */
+  if (stream) mail_close (stream);
+  syslog (LOG_INFO,"Logout user=%.80s host=%.80s",user ? user : "???",
+	  tcp_clienthost ());
+  return 0;			/* all done */
 }
 
 /* Clock interrupt
@@ -172,13 +188,14 @@ void main (int argc,char *argv[])
 
 void clkint ()
 {
-  char tmp[MAILTMPLEN];
-  puts ("- Autologout; idle for too long\015");
+  alarm (0);		/* disable all interrupts */
+  server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
+  fputs ("- Autologout; idle for too long\015\012",stdout);
   syslog (LOG_INFO,"Autologout user=%.80s host=%.80s",user ? user : "???",
-	  tcp_clienthost (tmp));
+	  tcp_clienthost ());
   fflush (stdout);		/* make sure output blatted */
-  if (critical) state = DONE;	/* badly host if in critical code */
-  else {			/* try to gracefully close the stream */
+  state = DONE;			/* mark state done in either case */
+  if (!critical) {		/* badly host if in critical code */
     if (stream && !stream->lock) mail_close (stream);
     stream = NIL;
     _exit (1);			/* die die die */
@@ -191,16 +208,20 @@ void clkint ()
 
 void kodint ()
 {
-  char tmp[MAILTMPLEN];
-  puts ("- Received Kiss of Death\015");
-  syslog (LOG_INFO,"Killed (lost mailbox lock) user=%.80s host=%.80s",
-	  user ? user : "???",tcp_clienthost (tmp));
-  fflush (stdout);		/* make sure output blatted */
-  if (critical) state = DONE;	/* badly host if in critical code */
-  else {			/* try to gracefully close the stream */
-    if (stream && !stream->lock) mail_close (stream);
-    stream = NIL;
-    _exit (1);			/* die die die */
+				/* only if in command wait */
+  if (idletime && ((time (0) - idletime) > KODTIMEOUT)) {
+    alarm (0);		/* disable all interrupts */
+    server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
+    fputs ("- Received Kiss of Death\015\012",stdout);
+    syslog (LOG_INFO,"Killed (lost mailbox lock) user=%.80s host=%.80s",
+	    user ? user : "???",tcp_clienthost ());
+    fflush (stdout);		/* make sure output blatted */
+    state = DONE;		/* mark state done in either case */
+    if (!critical) {		/* badly host if in critical code */
+      if (stream && !stream->lock) mail_close (stream);
+      stream = NIL;
+      _exit (1);		/* die die die */
+    }
   }
 }
 
@@ -210,11 +231,12 @@ void kodint ()
 
 void hupint ()
 {
-  char tmp[MAILTMPLEN];
+  alarm (0);		/* disable all interrupts */
+  server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
   syslog (LOG_INFO,"Hangup user=%.80s host=%.80s",user ? user : "???",
-	  tcp_clienthost (tmp));
-  if (critical) state = DONE;	/* badly host if in critical code */
-  else {			/* try to gracefully close the stream */
+	  tcp_clienthost ());
+  state = DONE;			/* mark state done in either case */
+  if (!critical) {		/* badly host if in critical code */
     if (stream && !stream->lock) mail_close (stream);
     stream = NIL;
     _exit (1);			/* die die die */
@@ -227,13 +249,14 @@ void hupint ()
 
 void trmint ()
 {
-  char tmp[MAILTMPLEN];
-  puts ("- Killed\015");
+  alarm (0);		/* disable all interrupts */
+  server_init (NIL,NIL,NIL,NIL,SIG_IGN,SIG_IGN,SIG_IGN,SIG_IGN);
+  fputs ("- Killed\015\012",stdout);
   syslog (LOG_INFO,"Killed user=%.80s host=%.80s",user ? user : "???",
-	  tcp_clienthost (tmp));
+	  tcp_clienthost ());
   fflush (stdout);		/* make sure output blatted */
-  if (critical) state = DONE;	/* badly host if in critical code */
-  else {			/* try to gracefully close the stream */
+  state = DONE;			/* mark state done in either case */
+  if (!critical) {		/* badly host if in critical code */
     if (stream && !stream->lock) mail_close (stream);
     stream = NIL;
     _exit (1);			/* die die die */
@@ -249,31 +272,35 @@ short c_helo (char *t,int argc,char *argv[])
 {
   char *s,*u,*p;
   char tmp[TMPLEN];
-  if (!(t && *t && (u = strtok (t," ")) && (p = strtok (NIL,"\015\012")))) {
-    puts ("- Missing user or password\015");
+  if ((!(t && *t && (u = strtok (t," ")) && (p = strtok (NIL,"\015\012")))) ||
+      (strlen (p) >= TMPLEN)) {	/* get user name and password */
+    fputs ("- Missing user or password\015\012",stdout);
     return DONE;
   }
 				/* copy password, handle quoting */
   for (s = tmp; *p; p++) *s++ = (*p == '\\') ? *++p : *p;
   *s = '\0';			/* tie off string */
   pass = cpystr (tmp);
-  if (s = strchr (u,':')) {	/* want remote mailbox? */
+#ifndef DISABLE_POP_PROXY
+				/* want remote mailbox? */
+  if ((s = strchr (u,':')) && anonymous_login (argc,argv)) {
     *s++ = '\0';		/* separate host name from user name */
     user = cpystr (s);		/* note user name */
     syslog (LOG_INFO,"IMAP login to host=%.80s user=%.80s host=%.80s",u,user,
-	    tcp_clienthost (tmp));
+	    tcp_clienthost ());
 				/* initially remote INBOX */
-    sprintf (tmp,"{%.128s}INBOX",u);
-    anonymous_login ();		/* try to become someone harmless */
+    sprintf (tmp,"{%.128s/user=%.128s}INBOX",u,user);
+				/* disable rimap just in case */
+    mail_parameters (NIL,SET_RSHTIMEOUT,0);
   }
-  else if (server_login (user = cpystr (u),pass,argc,argv)) {
-    syslog (LOG_INFO,"Login user=%.80s host=%.80s",user,tcp_clienthost (tmp));
+  else
+#endif
+  if (!s && server_login (user = cpystr (u),pass,argc,argv)) {
+    syslog (LOG_INFO,"Login user=%.80s host=%.80s",user,tcp_clienthost ());
     strcpy (tmp,"INBOX");	/* local; attempt login, select INBOX */
   }
   else {
-    puts ("- Bad login\015");
-    syslog (LOG_INFO,"Login failure user=%.80s host=%.80s",user,
-	    tcp_clienthost(tmp));
+    fputs ("- Bad login\015\012",stdout);
     return DONE;
   }
   return c_fold (tmp);		/* open default mailbox */
@@ -286,22 +313,32 @@ short c_helo (char *t,int argc,char *argv[])
 
 short c_fold (char *t)
 {
-  unsigned long i,j;
-  char *s,tmp[TMPLEN];
+  unsigned long i,j,flags;
+  char *s = NIL,tmp[2*TMPLEN];
+  NETMBX mb;
   if (!(t && *t)) {		/* make sure there's an argument */
-    puts ("- Missing mailbox name\015");
+    fputs ("- Missing mailbox name\015\012",stdout);
     return DONE;
   }
+  myusername_full (&flags);	/* get user type flags */
 				/* expunge old stream */
   if (stream && nmsgs) mail_expunge (stream);
   nmsgs = 0;			/* no more messages */
   if (msg) fs_give ((void **) &msg);
-				/* don't permit proxy to leave IMAP */
-  if (stream && stream->mailbox && (s = strchr (stream->mailbox,'}'))) {
-    strncpy (tmp,stream->mailbox,i = (++s - stream->mailbox));
-    strcpy (tmp+i,t);		/* append mailbox to initial spec */
-    t = tmp;
+#ifndef DISABLE_POP_PROXY
+  if (flags == MU_ANONYMOUS) {	/* don't permit proxy to leave IMAP */
+    if (stream) {		/* not first time */
+      if (!(stream->mailbox && (s = strchr (stream->mailbox,'}'))))
+	fatal ("bad previous mailbox name");
+      strncpy (tmp,stream->mailbox,i = (++s - stream->mailbox));
+      if (i >= TMPLEN) fatal ("ridiculous network prefix");
+      strcpy (tmp+i,t);		/* append mailbox to initial spec */
+      t = tmp;
+    }
+				/* must be net name first time */
+    else if (!mail_valid_net_parse (t,&mb)) fatal ("anonymous folder bogon");
   }
+#endif
 				/* open mailbox, note # of messages */
   if (j = (stream = mail_open (stream,t,NIL)) ? stream->nmsgs : 0) {
     sprintf (tmp,"1:%lu",j);	/* fetch fast information for all messages */
@@ -311,6 +348,12 @@ short c_fold (char *t)
     for (i = 1; i <= j; i++)	/* find undeleted messages, add to vector */
       if (!mail_elt (stream,i)->deleted) msg[++nmsgs] = i;
   }
+#ifndef DISABLE_POP_PROXY
+  if (!stream && (flags == MU_ANONYMOUS)) {
+    fputs ("- Bad login\015\012",stdout);
+    return DONE;
+  }
+#endif
   printf ("#%lu messages in %s\015\012",nmsgs,stream ? stream->mailbox :
 	  "<none>");
   return MBOX;
@@ -327,12 +370,12 @@ short c_read (char *t)
   if (t && *t) {		/* have a message number argument? */
 				/* validity check message number */
     if (((current = strtoul (t,NIL,10)) < 1) || (current > nmsgs)) {
-      puts ("- Invalid message number given to READ\015");
+      fputs ("- Invalid message number given to READ\015\012",stdout);
       return DONE;
     }
   }
   else if (current > nmsgs) {	/* at end of mailbox? */
-    puts ("=0 No more messages\015");
+    fputs ("=0 No more messages\015\012",stdout);
     return MBOX;
   }
 				/* set size if message valid and exists */
@@ -355,7 +398,7 @@ short c_read (char *t)
 short c_retr (char *t)
 {
   if (t) {			/* disallow argument */
-    puts ("- Bogus argument given to RETR\015");
+    fputs ("- Bogus argument given to RETR\015\012",stdout);
     return DONE;
   }
   if (size) {			/* message size valid? */
@@ -369,13 +412,13 @@ short c_retr (char *t)
       }
     }
     fputs (status,stdout);	/* yes, output message */
-    puts ("\015");		/* delimit header from text */
+    fputs ("\015\012",stdout);	/* delimit header from text */
     t = mail_fetch_text (stream,msg[current],NIL,&i,NIL);
     while (i) {			/* blat the text */
       j = fwrite (t,sizeof (char),i,stdout);
       if (i -= j) t += j;	/* advance to incomplete data */
     }
-    puts ("\015");		/* trailer to coddle PCNFS' NFSMAIL */
+    fputs ("\015\012",stdout);	/* trailer to coddle PCNFS' NFSMAIL */
   }
   else return DONE;		/* otherwise go away */
   return NEXT;
@@ -390,7 +433,7 @@ short c_acks (char *t)
 {
   char tmp[TMPLEN];
   if (t) {			/* disallow argument */
-    puts ("- Bogus argument given to ACKS\015");
+    fputs ("- Bogus argument given to ACKS\015\012",stdout);
     return DONE;
   }
 				/* mark message as seen */
@@ -409,7 +452,7 @@ short c_ackd (char *t)
 {
   char tmp[TMPLEN];
   if (t) {			/* disallow argument */
-    puts ("- Bogus argument given to ACKD\015");
+    fputs ("- Bogus argument given to ACKD\015\012",stdout);
     return DONE;
   }
 				/* mark message as seen and deleted */
@@ -428,7 +471,7 @@ short c_ackd (char *t)
 short c_nack (char *t)
 {
   if (t) {			/* disallow argument */
-    puts ("- Bogus argument given to NACK\015");
+    fputs ("- Bogus argument given to NACK\015\012",stdout);
     return DONE;
   }
   return c_read (NIL);		/* end message reading transaction */
@@ -467,7 +510,13 @@ void mm_exists (MAILSTREAM *stream,unsigned long number)
 
 void mm_expunged (MAILSTREAM *stream,unsigned long number)
 {
-  /* This isn't used */
+  if (state != DONE) {		/* ignore if closing */
+				/* this should never happen */
+    fputs ("- Mailbox expunged from under me!\015\012",stdout);
+    if (stream && !stream->lock) mail_close (stream);
+    stream = NIL;
+    _exit (1);
+  }
 }
 
 
@@ -562,8 +611,9 @@ void mm_dlog (char *string)
 void mm_login (NETMBX *mb,char *username,char *password,long trial)
 {
 				/* set user name */
-  strncpy (username,*mb->user ? mb->user : user,NETMAXUSER);
-  strncpy (password,pass,256);	/* and password */
+  strncpy (username,*mb->user ? mb->user : user,NETMAXUSER-1);
+  strncpy (password,pass,255);	/* and password */
+  username[NETMAXUSER] = password[255] = '\0';
 }
 
 /* About to enter critical code
@@ -595,14 +645,21 @@ void mm_nocritical (MAILSTREAM *stream)
 
 long mm_diskerror (MAILSTREAM *stream,long errcode,long serious)
 {
-  char tmp[MAILTMPLEN];
-  syslog (LOG_ALERT,
-	  "Retrying after disk error user=%.80s host=%.80s mbx=%.80s: %.80s",
-	  user,tcp_clienthost (tmp),
+  if (serious) {		/* try your damnest if clobberage likely */
+    syslog (LOG_ALERT,
+	    "Retrying after disk error user=%.80s host=%.80s mbx=%.80s: %.80s",
+	    user,tcp_clienthost (),
+	    (stream && stream->mailbox) ? stream->mailbox : "???",
+	    strerror (errcode));
+    alarm (0);			/* make damn sure timeout disabled */
+    sleep (60);			/* give it some time to clear up */
+    return NIL;
+  }
+  syslog (LOG_ALERT,"Fatal disk error user=%.80s host=%.80s mbx=%.80s: %.80s",
+	  user,tcp_clienthost (),
 	  (stream && stream->mailbox) ? stream->mailbox : "???",
 	  strerror (errcode));
-  sleep (5);			/* can't do much better than this! */
-  return NIL;
+  return T;
 }
 
 
